#!/usr/bin/env python
#
#    Pure python p7zr implementation
#    Copyright (C) 2019-2024 Hiroshi Miura
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
import hashlib
import io
from abc import ABC, abstractmethod
from typing import Optional, Union


class Py7zIO(ABC):

    @abstractmethod
    def write(self, s: bytes | bytearray) -> int:
        pass

    @abstractmethod
    def read(self, size: int | None = None) -> bytes:
        pass

    @abstractmethod
    def seek(self, offset: int, whence: int = 0) -> int:
        pass

    @abstractmethod
    def flush(self) -> None:
        pass

    @abstractmethod
    def size(self) -> int:
        pass

    def close(self) -> None:
        """Called by py7zr when decompression of an individual archive
        member completes. Override to release per-file resources or to
        process the just-decompressed payload before py7zr moves on to
        the next file. Default implementation is a no-op for backward
        compatibility with subclasses written before this hook existed."""
        return None


class HashIO(Py7zIO):
    def __init__(self, filename):
        self.filename = filename
        self.hash = hashlib.sha256()
        self._size: int = 0

    def write(self, s: bytes | bytearray) -> int:
        self._size += len(s)
        self.hash.update(s)
        return len(s)

    def read(self, length: int | None = None) -> bytes:
        return self.hash.digest()

    def seek(self, offset: int, whence: int = 0) -> int:
        return 0

    def flush(self) -> None:
        pass

    def size(self) -> int:
        return self._size


class Py7zBytesIO(Py7zIO):
    def __init__(self, filename: str, limit: int):
        self.filename = filename
        self.limit = limit
        self._buffer = io.BytesIO()

    def write(self, s: bytes | bytearray) -> int:
        if self.size() < self.limit:
            return self._buffer.write(s)
        else:
            return 0

    def read(self, size: int | None = None) -> bytes:
        return self._buffer.read(size)

    def seek(self, offset: int, whence: int = 0) -> int:
        return self._buffer.seek(offset, whence)

    def flush(self) -> None:
        return self._buffer.flush()

    def size(self) -> int:
        current_pos = self._buffer.tell()
        self._buffer.seek(0, io.SEEK_END)
        size = self._buffer.tell()
        self._buffer.seek(current_pos)
        return size


class WriterFactory(ABC):

    @abstractmethod
    def create(self, filename: str) -> Py7zIO:
        """request writer which compatible with BinaryIO, such as BytesIO object."""
        pass


class HashIOFactory(WriterFactory):
    def __init__(self):
        self.products = {}

    def create(self, filename: str) -> Py7zIO:
        product = HashIO(filename)
        self.products[filename] = product
        return product

    def get(self, filename: str) -> Py7zIO:
        return self.products[filename]


class BytesIOFactory(WriterFactory):

    def __init__(self, limit: int):
        self.limit = limit
        self.products: dict[str, Py7zBytesIO] = {}

    def create(self, filename: str) -> Py7zIO:
        product = Py7zBytesIO(filename, self.limit)
        self.products[filename] = product
        return product

    def get(self, filename):
        return self.products[filename]


class NullIOFactory(WriterFactory):
    def __init__(self):
        pass

    def create(self, filename: str) -> Py7zIO:
        return NullIO()


class MemIO:
    """pathlib.Path-like IO class to write memory"""

    def __init__(self, fname: str, factory: WriterFactory):
        self._buf: Py7zIO | None = None
        self._closed = True
        self.fname = fname
        self.factory = factory

    @property
    def mode(self) -> str:
        return self._mode

    @property
    def name(self) -> str:
        return self.fname

    @property
    def closed(self) -> bool:
        return self._closed

    def size(self) -> int:
        return self.__sizeof__()

    def __sizeof__(self) -> int:
        if self._buf is None:
            return -1
        return self._buf.size()

    def flush(self) -> None:
        if self._buf is not None:
            self._buf.flush()

    def write(self, s: bytes | bytearray) -> int:
        if self._buf is None:
            return -1
        return self._buf.write(s)

    def read(self, length: int | None = None) -> bytes:
        if self._buf is None:
            return b""
        if length is not None:
            return self._buf.read(length)
        else:
            return self._buf.read()

    def close(self) -> None:
        self._closed = True
        if self._buf is not None:
            self._buf.seek(0)
            self._buf.close()

    def seek(self, offset: int, whence: int = 0) -> int:
        if self._buf is None:
            return -1
        return self._buf.seek(offset, whence)

    def open(self, mode: str = "w"):
        self._mode = mode
        self._closed = False
        if self._buf is None:
            self._buf = self.factory.create(self.fname)
        else:
            self._buf.seek(0)
        return self

    @property
    def parent(self):
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.close()
        else:
            # extraction failed (e.g. CRC or decompression error); do not
            # signal completion through Py7zIO.close()
            self._closed = True


class NullIO(Py7zIO):
    """pathlib.Path-like IO class of /dev/null"""

    def __init__(self):
        pass

    def write(self, data):
        return len(data)

    def read(self, length=None):
        if length is not None:
            return bytes(length)
        else:
            return b""

    def close(self):
        pass

    def flush(self):
        pass

    def open(self, mode=None):
        return self

    @property
    def parent(self):
        return self

    def mkdir(self):
        return None

    def seek(self, offset: int, whence: int = 0) -> int:
        return offset

    def size(self):
        return 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class BufferOverflow(Exception):
    pass


class Buffer:
    def __init__(self, size: int = 16):
        self._buf = bytearray(size)
        self._buflen = 0
        self.view = memoryview(self._buf[0:0])

    def add(self, data: bytes | bytearray | memoryview):
        length = len(data)
        self._buf[self._buflen :] = data
        self._buflen += length
        self.view = memoryview(self._buf[0 : self._buflen])

    def reset(self) -> None:
        self._buflen = 0
        self.view = memoryview(self._buf[0:0])

    def set(self, data: bytes | bytearray | memoryview) -> None:
        length = len(data)
        self._buf[0:] = data
        self._buflen = length
        self.view = memoryview(self._buf[0:length])

    def get(self) -> bytearray:
        val = self._buf[: self._buflen]
        self.reset()
        return val

    def __len__(self) -> int:
        return self._buflen

    def __bytes__(self):
        return bytes(self._buf[0 : self._buflen])



#!/usr/bin/python -u
#
# p7zr library
#
# Copyright (c) 2019-2024 Hiroshi Miura <miurahr@linux.com>
# Copyright (c) 2004-2015 by Joachim Bauch, mail@joachim-bauch.de
# 7-Zip Copyright (C) 1999-2010 Igor Pavlov
# LZMA SDK Copyright (C) 1999-2010 Igor Pavlov
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
"""Read 7zip format archives."""

from __future__ import annotations

import collections.abc
import contextlib
import datetime
import errno
import functools
import io
import os
import pathlib
import queue
import re
import stat
import sys
import time
from collections.abc import Collection
from dataclasses import dataclass
from multiprocessing import Process
from shutil import ReadError
from threading import Thread
from typing import IO, TYPE_CHECKING, Any, BinaryIO, Protocol, TypedDict

import multivolumefile

from py7zr.archiveinfo import Folder, Header, SignatureHeader
from py7zr.callbacks import ExtractCallback
from py7zr.compressor import SupportedMethods, get_methods_names
from py7zr.exceptions import (
    AbsolutePathError,
    Bad7zFile,
    CrcError,
    DecompressionBombError,
    DecompressionError,
    InternalError,
    UnsupportedCompressionMethodError,
)
from py7zr.helpers import (
    ArchiveTimestamp,
    calculate_crc32,
    check_archive_path,
    filetime_to_dt,
    get_sanitized_output_path,
    is_path_valid,
    readlink,
    remove_trailing_slash,
)
from py7zr.io import MemIO, NullIO, WriterFactory
from py7zr.member import FILE_ATTRIBUTE_UNIX_EXTENSION, MemberType
from py7zr.properties import DEFAULT_FILTERS, FILTER_DEFLATE64, MAGIC_7Z, get_default_blocksize, get_memory_limit

if TYPE_CHECKING:
    from typing_extensions import NotRequired

if sys.platform.startswith("win"):
    import _winapi


class SupportsReadAndSeek(Protocol):
    def read(self, n: int = -1) -> bytes: ...
    def seek(self, offset: int, whence: int = 0) -> int: ...


class FileInfoDict(TypedDict):
    """
    File information dictionary structure for archive members.
    Contains filesystem metadata, timestamps, and platform-specific attributes.
    """

    # Core fields (always present)
    origin: pathlib.Path | None
    filename: str
    creationtime: ArchiveTimestamp
    lastwritetime: ArchiveTimestamp
    lastaccesstime: ArchiveTimestamp

    # Optional fields
    emptystream: NotRequired[bool]
    attributes: NotRequired[int]
    uncompressed: NotRequired[int]
    data: NotRequired[IO[Any]]
    readonly: NotRequired[bool]
    posix_mode: NotRequired[int | None]
    archivable: NotRequired[bool]
    is_directory: NotRequired[bool]


class ArchiveFile:
    """Represent each files metadata inside archive file.
    It holds file properties; filename, permissions, and type whether
    it is directory, link or normal file.

    Instances of the :class:`ArchiveFile` class are returned by iterating :attr:`files_list` of
    :class:`SevenZipFile` objects.
    Each object stores information about a single member of the 7z archive. Most of users use :meth:`extractall()`.

    The class also hold an archive parameter where file is exist in
    archive file folder(container)."""

    def __init__(self, id: int, file_info: FileInfoDict) -> None:
        self.id = id
        self._file_info = file_info

    def file_properties(self) -> FileInfoDict:
        """Return file properties as a hash object. Following keys are included: 'readonly', 'is_directory',
        'posix_mode', 'archivable', 'emptystream', 'filename', 'creationtime', 'lastaccesstime',
        'lastwritetime', 'attributes'
        """
        properties = self._file_info
        if properties is not None:
            properties["readonly"] = self.readonly
            properties["posix_mode"] = self.posix_mode
            properties["archivable"] = self.archivable
            properties["is_directory"] = self.is_directory
        return properties

    def _get_property(self, key: str) -> Any:
        return self._file_info.get(key)

    @property
    def origin(self) -> pathlib.Path:
        return self._get_property("origin")

    @property
    def folder(self) -> Folder:
        return self._get_property("folder")

    @property
    def filename(self) -> str:
        """return filename of archive file."""
        return self._get_property("filename")

    @property
    def emptystream(self) -> bool:
        """True if file is empty(0-byte file), otherwise False"""
        return self._get_property("emptystream")

    @property
    def uncompressed(self) -> int:
        return self._get_property("uncompressed")

    @property
    def compressed(self) -> int | None:
        """Compressed size"""
        return self._get_property("compressed")

    @property
    def crc32(self) -> int | None:
        """CRC of archived file(optional)"""
        return self._get_property("digest")

    def _test_attribute(self, target_bit: int) -> bool:
        attributes = self._get_property("attributes")
        if attributes is None:
            return False
        return attributes & target_bit == target_bit

    @property
    def archivable(self) -> bool:
        """File has a Windows `archive` flag."""
        if hasattr(stat, "FILE_ATTRIBUTE_ARCHIVE"):
            return self._test_attribute(getattr(stat, "FILE_ATTRIBUTE_ARCHIVE"))
        return False

    @property
    def is_directory(self) -> bool:
        """True if file is a directory, otherwise False."""
        if hasattr(stat, "FILE_ATTRIBUTE_DIRECTORY"):
            return self._test_attribute(getattr(stat, "FILE_ATTRIBUTE_DIRECTORY"))
        return False

    @property
    def is_file(self) -> bool:
        e = self._get_unix_extension()
        if e is not None:
            return stat.S_ISREG(e)
        return not (self.is_directory or self.is_symlink or self.is_junction or self.is_socket)

    @property
    def readonly(self) -> bool:
        """True if file is readonly, otherwise False."""
        if hasattr(stat, "FILE_ATTRIBUTE_READONLY"):
            return self._test_attribute(getattr(stat, "FILE_ATTRIBUTE_READONLY"))
        return False

    def _get_unix_extension(self) -> int | None:
        attributes = self._get_property("attributes")
        if self._test_attribute(FILE_ATTRIBUTE_UNIX_EXTENSION):
            return attributes >> 16
        return None

    def data(self) -> BinaryIO | None:
        return self._get_property("data")

    def has_strdata(self) -> bool:
        """True if file content is set by writestr() method otherwise False."""
        return "data" in self._file_info

    @property
    def is_symlink(self) -> bool:
        """True if file is a symbolic link, otherwise False."""
        e = self._get_unix_extension()
        if e is not None:
            return stat.S_ISLNK(e)
        if hasattr(stat, "FILE_ATTRIBUTE_REPARSE_POINT"):
            return self._test_attribute(getattr(stat, "FILE_ATTRIBUTE_REPARSE_POINT"))
        return False

    @property
    def is_junction(self) -> bool:
        """True if file is a junction/reparse point on windows, otherwise False."""
        if hasattr(stat, "FILE_ATTRIBUTE_REPARSE_POINT"):
            return self._test_attribute(
                getattr(stat, "FILE_ATTRIBUTE_REPARSE_POINT") | getattr(stat, "FILE_ATTRIBUTE_DIRECTORY")
            )
        return False

    @property
    def is_socket(self) -> bool:
        """True if file is a socket, otherwise False."""
        e = self._get_unix_extension()
        if e is not None:
            return stat.S_ISSOCK(e)
        return False

    @property
    def lastwritetime(self) -> ArchiveTimestamp | None:
        """Return last written timestamp of a file."""
        return self._get_property("lastwritetime")

    @property
    def posix_mode(self) -> int | None:
        """
        posix mode when a member has a unix extension property, or None
        :return: Return file stat mode can be set by os.chmod()
        """
        e = self._get_unix_extension()
        if e is not None:
            return stat.S_IMODE(e)
        return None

    @property
    def st_fmt(self) -> int | None:
        """
        :return: Return the portion of the file mode that describes the file type
        """
        e = self._get_unix_extension()
        if e is not None:
            return stat.S_IFMT(e)
        return None


class ArchiveFileList(collections.abc.Iterable[ArchiveFile]):
    """Iterable container of ArchiveFile."""

    def __init__(self, offset: int = 0):
        self.files_list: list[FileInfoDict] = []
        self.index = 0
        self.offset = offset

    def append(self, file_info: FileInfoDict) -> None:
        self.files_list.append(file_info)

    def __len__(self) -> int:
        return len(self.files_list)

    def __iter__(self) -> ArchiveFileListIterator:
        return ArchiveFileListIterator(self)

    def __getitem__(self, index: int) -> ArchiveFile:
        if index > len(self.files_list):
            raise IndexError
        if index < 0:
            raise IndexError
        res = ArchiveFile(index + self.offset, self.files_list[index])
        return res


class ArchiveFileListIterator(collections.abc.Iterator[ArchiveFile]):
    def __init__(self, archive_file_list: ArchiveFileList):
        self._archive_file_list = archive_file_list
        self._index = 0

    def __next__(self) -> ArchiveFile:
        if self._index == len(self._archive_file_list):
            raise StopIteration
        res = self._archive_file_list[self._index]
        self._index += 1
        return res


# ------------------
# Exported Classes
# ------------------
@dataclass
class ArchiveInfo:
    """Hold archive information"""

    filename: str
    stat: os.stat_result
    header_size: int
    method_names: list[str]
    solid: bool
    blocks: int
    uncompressed: int


@dataclass
class FileInfo:
    """Hold archived file information."""

    filename: str
    compressed: int | None
    uncompressed: int
    archivable: bool
    is_directory: bool
    is_file: bool
    is_symlink: bool
    creationtime: datetime.datetime | None
    crc32: int | None

    def __post_init__(self) -> None:
        # Prevent ambiguous file type states.
        # A file can't simultaneously be a directory, a regular file, or a symlink,
        # but it’s allowed to be none of these (e.g., a junction or a socket).
        flags = self.is_directory + self.is_file + self.is_symlink
        if flags > 1:
            raise ValueError(
                f"At most one of is_directory, is_file, or is_symlink can be True; "
                f"got is_directory={self.is_directory}, is_file={self.is_file}, is_symlink={self.is_symlink}"
            )


class SevenZipFile(contextlib.AbstractContextManager):
    """The SevenZipFile Class provides an interface to 7z archives."""

    def __init__(
        self,
        file: IO[bytes] | str | pathlib.Path,
        mode: str = "r",
        *,
        filters: list[dict[str, int]] | None = None,
        dereference=False,
        password: str | None = None,
        header_encryption: bool = False,
        blocksize: int | None = None,
        mp: bool = False,
        max_extract_size: int | None = None,
    ) -> None:
        # check invalid mode.
        if mode not in ("r", "w", "x", "a"):
            raise ValueError("SevenZipFile requires mode 'r', 'w', 'x', or 'a'")
        self.fp: IO[bytes]
        self.mp = mp
        self.password_protected = password is not None
        self.max_extract_size = max_extract_size
        if blocksize:
            self._block_size = blocksize
        else:
            self._block_size = get_default_blocksize()

        # https://github.com/python/cpython/blob/b5e142ba7c2063efe9bb8065c3b0bad33e2a9afa/Lib/zipfile/__init__.py#L1350
        # Check if we were passed a file-like object or not
        if isinstance(file, os.PathLike):
            file = os.fspath(file)
        if isinstance(file, str):
            # No, it's a filename
            self._filePassed = False
            self.filename: str | None = file
            modeDict = {
                "r": "rb",
                "w": "w+b",
                "x": "x+b",
                "a": "r+b",
                "r+b": "w+b",
                "w+b": "wb",
                "x+b": "xb",
            }
            filemode = modeDict[mode]

            while True:
                try:
                    self.fp = open(file, filemode)
                except OSError:
                    if filemode in modeDict:
                        filemode = modeDict[filemode]
                        continue
                    raise
                break
            self.mode = mode
        elif isinstance(file, multivolumefile.MultiVolume):
            self._filePassed = True
            self.fp = file
            self.filename = None
            self.mode = mode  # noqa
        elif isinstance(file, io.IOBase):
            self._filePassed = True
            self.fp = file
            self.filename = getattr(file, "name", None)
            self.mode = mode  # noqa
        else:
            raise TypeError(f"invalid file: {type(file)}")
        self.encoded_header_mode = True
        self.header_encryption = header_encryption
        try:
            if mode == "r":
                self._real_get_contents(password)
                self.fp.seek(self.afterheader)  # seek into start of payload and prepare worker to extract
                self.worker = Worker(self.files, self.afterheader, self.header, self.mp)
            elif mode == "w":
                self._prepare_write(filters, password)
            elif mode == "x":
                self._prepare_write(filters, password)
            elif mode == "a":
                try:
                    # Append if it's an existing 7zip file
                    self._real_get_contents(password)
                    self._prepare_append(filters, password)
                except Bad7zFile:
                    # Not an existing 7zip file, write instead
                    self._prepare_write(filters, password)
            else:
                raise ValueError("Mode must be 'r', 'w', 'x', or 'a'")  # never come here
        except Exception as e:
            self._fpclose()
            raise e
        self.dereference = dereference
        self.reporterd: Thread | None = None
        self.q: queue.Queue[Any] = queue.Queue()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _fpclose(self) -> None:
        if not self._filePassed:
            self.fp.close()

    def _real_get_contents(self, password) -> None:
        if not self._check_7zfile(self.fp):
            raise Bad7zFile("not a 7z file")
        self.sig_header = SignatureHeader.retrieve(self.fp)  # type: ignore[arg-type]
        self.afterheader: int = self.fp.tell()
        self.fp.seek(self.sig_header.nextheaderofs, os.SEEK_CUR)
        buffer = io.BytesIO(self.fp.read(self.sig_header.nextheadersize))
        if self.sig_header.nextheadercrc != calculate_crc32(buffer.getvalue()):
            raise Bad7zFile("invalid header data")
        header = Header.retrieve(self.fp, buffer, self.afterheader, password)  # type: ignore[arg-type]
        if header is None:
            return
        header.size += 32 + self.sig_header.nextheadersize
        self.header = header
        buffer.close()
        self.files = ArchiveFileList()
        if getattr(self.header, "files_info", None) is None:
            return
        # Initialize references for convenience
        if hasattr(self.header, "main_streams") and self.header.main_streams is not None:
            folders = self.header.main_streams.unpackinfo.folders
            for folder in folders:
                folder.password = password
            packinfo = self.header.main_streams.packinfo
            packsizes = packinfo.packsizes
            subinfo = self.header.main_streams.substreamsinfo
            if subinfo is not None and subinfo.unpacksizes is not None:
                unpacksizes = subinfo.unpacksizes
            else:
                unpacksizes = [x.unpacksizes[-1] for x in folders]
        else:
            subinfo = None
            folders = None
            packinfo = None
            packsizes = []
            unpacksizes = [0]

        pstat = self.ParseStatus()
        pstat.src_pos = self.afterheader
        file_in_solid = 0

        for file_id, file_info in enumerate(self.header.files_info.files):
            if not file_info["emptystream"] and folders is not None:
                folder = folders[pstat.folder]
                numinstreams = max([coder.get("numinstreams", 1) for coder in folder.coders])
                maxsize, compressed, uncompressed, packsize, solid = self._get_fileinfo_sizes(
                    pstat, subinfo, packinfo, folder, packsizes, unpacksizes, file_in_solid, numinstreams
                )
                pstat.input += 1
                folder.solid = solid
                file_info["folder"] = folder
                file_info["maxsize"] = maxsize
                file_info["compressed"] = compressed
                file_info["uncompressed"] = uncompressed
                file_info["packsizes"] = packsize
                if subinfo.digestsdefined[pstat.outstreams]:
                    file_info["digest"] = subinfo.digests[pstat.outstreams]
                if folder is None:
                    pstat.src_pos += file_info["compressed"]
                else:
                    if folder.solid:
                        file_in_solid += 1
                    pstat.outstreams += 1
                    if folder.files is None:
                        folder.files = ArchiveFileList(offset=file_id)
                    folder.files.append(file_info)
                    if pstat.input >= subinfo.num_unpackstreams_folders[pstat.folder]:
                        file_in_solid = 0
                        pstat.src_pos += sum(packinfo.packsizes[pstat.stream : pstat.stream + numinstreams])
                        pstat.folder += 1
                        pstat.stream += numinstreams
                        pstat.input = 0
            else:
                file_info["folder"] = None
                file_info["maxsize"] = 0
                file_info["compressed"] = 0
                file_info["uncompressed"] = 0
                file_info["packsizes"] = [0]

            if "filename" not in file_info:
                # compressed file is stored without a name, generate one
                try:
                    basefilename = self.filename
                except AttributeError:
                    # 7z archive file doesn't have a name
                    file_info["filename"] = "contents"
                else:
                    if basefilename is not None:
                        fn, ext = os.path.splitext(os.path.basename(basefilename))
                        file_info["filename"] = fn
                    else:
                        file_info["filename"] = "contents"
            self.files.append(file_info)
        if not self.password_protected and self.header.main_streams is not None:
            # Check specified coders have a crypt method or not.
            self.password_protected = any(
                [SupportedMethods.needs_password(folder.coders) for folder in self.header.main_streams.unpackinfo.folders]
            )

    def _extract(
        self,
        path: Any | None = None,
        targets: Collection[str] | None = None,
        callback: ExtractCallback | None = None,
        recursive: bool | None = False,
        writer_factory: WriterFactory | None = None,
    ) -> None:
        if callback is None:
            pass
        elif isinstance(callback, ExtractCallback):
            self.reporterd = Thread(target=self.reporter, args=(callback,), daemon=True)
            self.reporterd.start()
        else:
            raise ValueError("Callback specified is not an instance of subclass of py7zr.callbacks.ExtractCallback class")
        target_files: list[tuple[pathlib.Path, FileInfoDict]] = []
        target_dirs: list[pathlib.Path] = []
        if path is not None:
            if isinstance(path, str):
                path = pathlib.Path(path)
            try:
                if not path.exists():
                    path.mkdir(parents=True)
                else:
                    pass
            except OSError as e:
                if e.errno == errno.EEXIST and path.is_dir():
                    pass
                else:
                    raise e
        if targets is not None:
            # faster lookups
            targets = set(targets)
        fnames: dict[str, int] = {}  # check duplicated filename in one archive?
        self.q.put(("pre", None, None))
        for f in self.files:
            if targets is not None and recursive is False:
                if f.filename not in targets:
                    self.worker.register_filelike(f.id, None)
                    continue
            elif targets is not None and recursive is True:
                if f.filename not in targets and not any([f.filename.startswith(target) for target in targets]):
                    self.worker.register_filelike(f.id, None)
                    continue

            # When archive has a multiple files which have same name
            # To guarantee order of archive, multi-thread decompression becomes off.
            # Currently always overwrite by latter archives.
            # TODO: provide option to select overwrite or skip.
            if f.filename not in fnames:
                outname = f.filename
                fnames[f.filename] = 0
            else:
                outname = f"{f.filename}_{fnames[f.filename]}"
                fnames[f.filename] += 1
            if path is None or path.is_absolute():
                outfilename = get_sanitized_output_path(outname, path)
            else:
                outfilename = get_sanitized_output_path(outname, pathlib.Path(os.getcwd()).joinpath(path))
            if writer_factory is not None:
                if f.is_directory or f.is_socket:
                    # ignore special files and directories
                    pass
                else:
                    fname = outfilename.as_posix()
                    self.worker.register_filelike(f.id, MemIO(fname, writer_factory))
            elif f.is_directory:
                if not outfilename.exists():
                    target_dirs.append(outfilename)
                    target_files.append((outfilename, f.file_properties()))
                else:
                    pass
            elif f.is_socket:
                pass  # TODO: implement me.
            elif f.is_symlink or f.is_junction:
                self.worker.register_filelike(f.id, outfilename)
            else:
                self.worker.register_filelike(f.id, outfilename)
                target_files.append((outfilename, f.file_properties()))
        for target_dir in sorted(target_dirs):
            try:
                target_dir.mkdir(parents=True)
            except FileExistsError:
                if target_dir.is_dir():
                    pass
                elif target_dir.is_file():
                    raise DecompressionError(f"Directory {target_dir} is existed as a normal file.")
                else:
                    raise DecompressionError(f"Directory {target_dir} making fails on unknown condition.")

        self.worker.max_extract_size = self.max_extract_size
        if callback is not None:
            self.worker.extract(
                self.fp,
                path,
                parallel=(not self.password_protected and not self._filePassed),
                q=self.q,
            )
        else:
            self.worker.extract(
                self.fp,
                path,
                parallel=(not self.password_protected and not self._filePassed),
            )

        self.q.put(("post", None, None))
        # early return when dict specified
        if writer_factory is not None:
            return
        # set file properties
        for outfilename, properties in target_files:
            # mtime
            lastwritetime = properties.get("lastwritetime")
            if lastwritetime is not None:
                lastmodified = ArchiveTimestamp(lastwritetime).totimestamp()
                os.utime(str(outfilename), times=(lastmodified, lastmodified))
            if os.name == "posix":
                st_mode = properties["posix_mode"]
                if st_mode is not None:
                    outfilename.chmod(st_mode)
                    continue
            # fallback: only set readonly if specified
            if properties["readonly"] and not properties["is_directory"]:
                ro_mask = 0o777 ^ (stat.S_IWRITE | stat.S_IWGRP | stat.S_IWOTH)
                outfilename.chmod(outfilename.stat().st_mode & ro_mask)
        return None

    def _prepare_append(self, filters, password):
        if password is not None and filters is None:
            filters = DEFAULT_FILTERS.ENCRYPTED_ARCHIVE_FILTER
        elif filters is None:
            filters = DEFAULT_FILTERS.ARCHIVE_FILTER
        else:
            for f in filters:
                if f["id"] == FILTER_DEFLATE64:
                    raise UnsupportedCompressionMethodError(filters, "Compression with deflate64 is not supported.")
        self.header.filters = filters
        self.header.password = password
        if self.header.main_streams is not None:
            pos = self.afterheader + self.header.main_streams.packinfo.packpositions[-1]
        else:
            pos = self.afterheader
        self.fp.seek(pos)
        self.worker = Worker(self.files, pos, self.header, self.mp)

    def _prepare_write(self, filters, password):
        if password is not None and filters is None:
            filters = DEFAULT_FILTERS.ENCRYPTED_ARCHIVE_FILTER
        elif filters is None:
            filters = DEFAULT_FILTERS.ARCHIVE_FILTER
        self.files = ArchiveFileList()
        self.sig_header = SignatureHeader()
        self.sig_header._write_skeleton(self.fp)
        self.afterheader = self.fp.tell()
        self.header = Header.build_header(filters, password)
        self.fp.seek(self.afterheader)
        self.worker = Worker(self.files, self.afterheader, self.header, self.mp)

    def _write_flush(self):
        if self.header is not None:
            if self.header._initialized:
                folder = self.header.main_streams.unpackinfo.folders[-1]
                self.worker.flush_archive(self.fp, folder)
            self._write_header()

    def _write_header(self):
        """Write header and update signature header."""
        header_pos, header_len, header_crc = self.header.write(
            self.fp,
            self.afterheader,
            encoded=self.encoded_header_mode,
            encrypted=self.header_encryption,
        )
        self.sig_header.nextheaderofs = header_pos - self.afterheader
        self.sig_header.calccrc(header_len, header_crc)
        self.sig_header.write(self.fp)

    def _writeall(self, path, arcname):
        try:
            if path.is_symlink() and not self.dereference:
                self.write(path, arcname)
            elif path.is_file():
                self.write(path, arcname)
            elif path.is_dir():
                if not path.samefile("."):
                    self.write(path, arcname)
                for nm in sorted(os.listdir(str(path))):
                    arc = os.path.join(arcname, nm) if arcname is not None else None
                    self._writeall(path.joinpath(nm), arc)
            else:
                return  # pathlib ignores ELOOP and return False for is_*().
        except OSError as ose:
            if self.dereference and ose.errno in [errno.ELOOP]:
                return  # ignore ELOOP here, this resulted to stop looped symlink reference.
            elif self.dereference and sys.platform == "win32" and ose.errno in [errno.ENOENT]:
                return  # ignore ENOENT which is happened when a case of ELOOP on windows.
            else:
                raise

    class ParseStatus:
        def __init__(self, src_pos=0):
            self.src_pos = src_pos
            self.folder = 0  # 7zip folder where target stored
            self.outstreams = 0  # output stream count
            self.input = 0  # unpack stream count in each folder
            self.stream = 0  # target input stream position

    def _get_fileinfo_sizes(
        self,
        pstat,
        subinfo,
        packinfo,
        folder,
        packsizes,
        unpacksizes,
        file_in_solid,
        numinstreams,
    ):
        if pstat.input == 0:
            folder.solid = subinfo.num_unpackstreams_folders[pstat.folder] > 1
        maxsize: int | None = (folder.solid and packinfo.packsizes[pstat.stream]) or None
        uncompressed: int = unpacksizes[pstat.outstreams]
        if file_in_solid > 0:
            compressed: int | None = None
        elif pstat.stream < len(packsizes):  # file is compressed
            compressed = packsizes[pstat.stream]
        else:  # file is not compressed
            compressed = uncompressed
        packsize: int = packsizes[pstat.stream : pstat.stream + numinstreams]
        return maxsize, compressed, uncompressed, packsize, folder.solid

    def set_encoded_header_mode(self, mode: bool) -> None:
        if mode:
            self.encoded_header_mode = True
        else:
            self.encoded_header_mode = False
            self.header_encryption = False

    def set_encrypted_header(self, mode: bool) -> None:
        if mode:
            self.encoded_header_mode = True
            self.header_encryption = True
        else:
            self.header_encryption = False

    @staticmethod
    def _check_7zfile(fp: SupportsReadAndSeek) -> bool:
        try:
            magic_len = len(MAGIC_7Z)
            result = MAGIC_7Z == fp.read(magic_len)[:magic_len]
            fp.seek(-magic_len, 1)
            return result
        except OSError:
            # A new empty file raises OSError
            return False

    def _get_method_names(self) -> list[str]:
        try:
            return get_methods_names([folder.coders for folder in self.header.main_streams.unpackinfo.folders])
        except KeyError:
            raise UnsupportedCompressionMethodError(self.header.main_streams.unpackinfo.folders, "Unknown method")

    def _read_digest(self, pos: int, size: int) -> int:
        self.fp.seek(pos)
        remaining_size = size
        digest = 0
        while remaining_size > 0:
            block = min(self._block_size, remaining_size)
            digest = calculate_crc32(self.fp.read(block), digest)
            remaining_size -= block
        return digest

    def _is_solid(self):
        for f in self.header.main_streams.substreamsinfo.num_unpackstreams_folders:
            if f > 1:
                return True
        return False

    def _var_release(self):
        self._dict = {}
        self.worker.close()
        del self.worker
        del self.files
        del self.header
        del self.sig_header

    @staticmethod
    def _make_file_info(target: pathlib.Path, arcname: str | None = None, dereference: bool = False) -> FileInfoDict:
        origin = target
        filename = pathlib.Path(arcname).as_posix() if arcname else target.as_posix()
        target = target.resolve() if dereference else target
        fstat = target.lstat()
        f = FileInfoDict(
            origin=origin,
            filename=filename,
            creationtime=ArchiveTimestamp.from_datetime(fstat.st_ctime),
            lastwritetime=ArchiveTimestamp.from_datetime(fstat.st_mtime),
            lastaccesstime=ArchiveTimestamp.from_datetime(fstat.st_atime),
        )

        if target.is_symlink():
            f["emptystream"] = False
            f["attributes"] = MemberType.SYMLINK.attributes(fstat)
            # TODO: handle junctions
        elif target.is_dir():
            f["emptystream"] = True
            f["attributes"] = MemberType.DIRECTORY.attributes(fstat)
        elif target.is_file():
            f["emptystream"] = False
            f["uncompressed"] = fstat.st_size
            f["attributes"] = MemberType.FILE.attributes(fstat)

        return f

    @staticmethod
    def _make_file_info_from_name(bio: IO[Any], size: int, arcname: str) -> FileInfoDict:
        return FileInfoDict(
            origin=None,
            data=bio,
            filename=pathlib.Path(arcname).as_posix(),
            uncompressed=size,
            emptystream=False,
            attributes=MemberType.FILE.attributes(),
            creationtime=ArchiveTimestamp.from_now(),
            lastwritetime=ArchiveTimestamp.from_now(),
            lastaccesstime=ArchiveTimestamp.from_now(),
        )

    @staticmethod
    def _sanitize_archive_arcname(arcname):
        if isinstance(arcname, str):
            path = arcname
        else:
            path = str(arcname)
        # Strip leading / (tar's directory separator) from filenames.
        # Include os.sep (target OS directory separator) as well.
        if path.startswith(("/", os.sep)):
            path = path.lstrip("/" + os.sep)
        if re.match("^[a-zA-Z]:", path):
            path = path[2:]
            # strip again
            if path.startswith(("/", os.sep)):
                path = path.lstrip("/" + os.sep)
        if os.path.isabs(path) or re.match("^[a-zA-Z]:", path):
            # Path is absolute even after stripping.
            raise AbsolutePathError(arcname)
        return path

    @staticmethod
    def _is_none_or_collection(t):
        if t is None:
            return True
        if isinstance(t, str):
            return False
        if isinstance(t, tuple):
            return False
        if isinstance(t, list) or isinstance(t, set):
            return True
        return False

    # --------------------------------------------------------------------------
    # The public methods which SevenZipFile provides:
    def getnames(self) -> list[str]:
        """Return the members of the archive as a list of their names. It has
        the same order as the list returned by getmembers().
        """
        return self.namelist()

    def namelist(self) -> list[str]:
        """Return a list of archive members by name."""
        return list(x.filename for x in self.files)

    def getinfo(self, name: str) -> FileInfo:
        """Return a :class:`FileInfo` object with information about the archive member *name*.
        Calling :meth:`getinfo()` for a name not currently contained in the archive will raise a :exc:`KeyError`."""
        # For interoperability with ZipFile
        name = remove_trailing_slash(name)

        try:
            sevenzipinfo = next(member for member in self.list() if member.filename == name)
        except StopIteration:
            # ZipFile and TarFile raise KeyError if the named member is not found
            # So for consistency, we'll also raise KeyError here
            raise KeyError(f"'{name}' not found in archive.") from None

        return sevenzipinfo

    def archiveinfo(self) -> ArchiveInfo:
        total_uncompressed = functools.reduce(lambda x, y: x + y, [f.uncompressed for f in self.files])
        if isinstance(self.fp, multivolumefile.MultiVolume):
            fname = self.fp.name
            fstat = self.fp.stat()
        else:
            fname = self.filename
            assert fname is not None
            fstat = os.stat(fname)
        return ArchiveInfo(
            fname,
            fstat,
            self.header.size,
            self._get_method_names(),
            self._is_solid(),
            len(self.header.main_streams.unpackinfo.folders),
            total_uncompressed,
        )

    def needs_password(self) -> bool:
        return self.password_protected

    def list(self) -> list[FileInfo]:
        """Returns contents information"""
        alist: list[FileInfo] = []
        lastmodified: datetime.datetime | None = None
        for f in self.files:
            if f.lastwritetime is not None:
                lastmodified = filetime_to_dt(f.lastwritetime)
            alist.append(
                FileInfo(
                    filename=f.filename,
                    compressed=f.compressed,
                    uncompressed=f.uncompressed,
                    archivable=f.archivable,
                    is_file=f.is_file,
                    is_directory=f.is_directory,
                    is_symlink=f.is_symlink,
                    creationtime=lastmodified,
                    crc32=f.crc32,
                )
            )
        return alist

    def extractall(
        self,
        path: Any | None = None,
        *,
        callback: ExtractCallback | None = None,
        factory: WriterFactory | None = None,
    ) -> None:
        """Extract all members from the archive to the current working
        directory and set owner, modification time and permissions on
        directories afterward. ``path`` specifies a different directory
        to extract to.
        """
        self._extract(path=path, callback=callback, writer_factory=factory)

    def extract(
        self,
        path: Any | None = None,
        targets: Collection[str] | None = None,
        recursive: bool | None = False,
        *,
        callback: ExtractCallback | None = None,
        factory: WriterFactory | None = None,
    ) -> None:
        if not self._is_none_or_collection(targets):
            raise TypeError("Wrong argument type given.")
        # For interoperability with ZipFile, we strip any trailing slashes
        # This also matches the behavior of TarFile
        if targets is not None:
            targets = [remove_trailing_slash(target) for target in targets]
        self._extract(path, targets, recursive=recursive, callback=callback, writer_factory=factory)

    def reporter(self, callback: ExtractCallback) -> None:
        while True:
            try:
                item: tuple[str, str, str] | None = self.q.get(timeout=1)
            except queue.Empty:
                pass
            else:
                if item is None:
                    break
                elif item[0] == "s":
                    callback.report_start(item[1], item[2])
                elif item[0] == "u":
                    callback.report_update(item[2])
                elif item[0] == "e":
                    callback.report_end(item[1], item[2])
                elif item[0] == "pre":
                    callback.report_start_preparation()
                elif item[0] == "post":
                    callback.report_postprocess()
                elif item[0] == "w":
                    callback.report_warning(item[1])
                else:
                    pass
                self.q.task_done()

    def writeall(
        self,
        path: pathlib.Path | str,
        arcname: str | None = None,
    ) -> None:
        """Write files in target path into archive."""
        if isinstance(path, str):
            path = pathlib.Path(path)
        if not path.exists():
            raise ValueError("specified path does not exist.")
        if path.is_dir() or path.is_file():
            self._writeall(path, arcname)
        else:
            raise ValueError("specified path is not a directory or a file")

    def write(
        self,
        file: pathlib.Path | str,
        arcname: str | None = None,
    ) -> None:
        """Write single target file into archive."""
        if not isinstance(file, str) and not isinstance(file, pathlib.Path):
            raise ValueError("Unsupported file type.")
        if arcname is None:
            arcname = self._sanitize_archive_arcname(file)
        else:
            arcname = self._sanitize_archive_arcname(arcname)
        if isinstance(file, str):
            path = pathlib.Path(file)
        else:
            path = file
        folder = self.header.initialize()
        file_info = self._make_file_info(path, arcname, self.dereference)
        self.header.files_info.files.append(file_info)
        self.header.files_info.emptyfiles.append(file_info["emptystream"])
        self.files.append(file_info)
        self.worker.archive(self.fp, self.files, folder, deref=self.dereference)

    def writef(self, bio: IO[Any], arcname: str) -> None:
        if not check_archive_path(arcname):
            raise ValueError(f"Specified path is bad: {arcname}")
        return self._writef(bio, arcname)

    def _writef(self, bio: IO[Any], arcname: str) -> None:
        # Check for null byte injection - 7z uses null-terminated strings in headers
        if "\\x00" in arcname:
            raise ValueError(f"Filename contains null byte: {arcname}")
        if isinstance(bio, io.BytesIO):
            size = bio.getbuffer().nbytes
        elif isinstance(bio, io.TextIOBase):
            # First check whether is it Text?
            raise ValueError("Unsupported file object type: please open file with Binary mode.")
        elif isinstance(bio, io.BufferedIOBase):
            # come here when subtype of io.BufferedIOBase that don't have __sizeof__ (eg. pypy)
            # alternative for `size = bio.__sizeof__()`
            current = bio.tell()
            bio.seek(0, os.SEEK_END)
            last = bio.tell()
            bio.seek(current, os.SEEK_SET)
            size = last - current
        elif isinstance(bio, MemIO):
            size = bio.__sizeof__()
        else:
            raise ValueError("Wrong argument passed for argument bio.")
        if size >= 0:
            folder = self.header.initialize()
            file_info = self._make_file_info_from_name(bio, size, arcname)
            self.header.files_info.files.append(file_info)
            self.header.files_info.emptyfiles.append(file_info["emptystream"])
            self.files.append(file_info)
            self.worker.archive(self.fp, self.files, folder, deref=False)
        else:
            file_info = self._make_file_info_from_name(bio, size, arcname)
            self.header.files_info.files.append(file_info)
            self.header.files_info.emptyfiles.append(file_info["emptystream"])
            self.files.append(file_info)

    def writestr(
        self,
        data: str | bytes | bytearray | memoryview,
        arcname: str,
    ) -> None:
        if not check_archive_path(arcname):
            raise ValueError(f"Specified path is bad: {arcname}")
        return self._writestr(data, arcname)

    def _writestr(
        self,
        data: str | bytes | bytearray | memoryview,
        arcname: str,
    ) -> None:
        if not isinstance(arcname, str):
            raise ValueError("Unsupported arcname")
        if isinstance(data, str):
            self._writef(io.BytesIO(data.encode("UTF-8")), arcname)
        elif isinstance(data, bytes) or isinstance(data, bytearray) or isinstance(data, memoryview):
            self._writef(io.BytesIO(bytes(data)), arcname)
        else:
            raise ValueError("Unsupported data type.")

    def close(self) -> None:
        """Flush all the data into archive and close it.
        When close py7zr start reading target and writing actual archive file.
        """
        if "r" in self.mode:
            if self.reporterd is not None:
                self.q.put_nowait(None)
                self.reporterd.join(1)
                if self.reporterd.is_alive():
                    raise InternalError("Progress report thread terminate error.")
                self.reporterd = None
        else:  # "w" | "x" | "a" in self.mode
            self._write_flush()

        self._fpclose()
        self._var_release()

    def reset(self) -> None:
        """
        When read mode, it reset file pointer, decompress worker and decompressor
        """
        if self.mode == "r":
            self.fp.seek(self.afterheader)
            self.worker = Worker(self.files, self.afterheader, self.header, self.mp)
            if self.header.main_streams is not None and self.header.main_streams.unpackinfo.numfolders > 0:
                for i, folder in enumerate(self.header.main_streams.unpackinfo.folders):
                    folder.decompressor = None

    def test(self) -> bool | None:
        self.fp.seek(self.afterheader)
        self.worker = Worker(self.files, self.afterheader, self.header, self.mp)
        crcs: list[int] | None = self.header.main_streams.packinfo.crcs
        if crcs is None or len(crcs) == 0:
            return None
        packpos = self.afterheader + self.header.main_streams.packinfo.packpos
        packsizes = self.header.main_streams.packinfo.packsizes
        digestdefined = self.header.main_streams.packinfo.digestdefined
        j = 0
        for i, d in enumerate(digestdefined):
            if d:
                if self._read_digest(packpos, packsizes[i]) != crcs[j]:
                    return False
                j += 1
            packpos += packsizes[i]
        return True

    def testzip(self) -> str | None:
        self.fp.seek(self.afterheader)
        self.worker = Worker(self.files, self.afterheader, self.header, self.mp)
        for f in self.files:
            self.worker.register_filelike(f.id, None)
        try:
            self.worker.extract(
                self.fp, None, parallel=(not self.password_protected), skip_notarget=False
            )  # TODO: print progress
        except CrcError as crce:
            return crce.args[2]
        else:
            return None


# --------------------
# exported functions
# --------------------
def is_7zfile(file: SupportsReadAndSeek | IO[bytes] | str | os.PathLike[str]) -> bool:
    """Quickly see if a file is a 7Z file by checking the magic number.
    The file argument may be a filename or file-like object too.
    """
    if hasattr(file, "read") and hasattr(file, "seek"):
        return SevenZipFile._check_7zfile(file)  # type: ignore

    try:
        if isinstance(file, str):
            with open(file, mode="rb") as fp:
                return SevenZipFile._check_7zfile(fp)

        if isinstance(file, os.PathLike):
            with open(os.fspath(file), mode="rb") as fp:
                return SevenZipFile._check_7zfile(fp)
    except FileNotFoundError:
        return False

    msg = (
        f"Invalid 'file' argument: Expected a path-like object " f"or a binary file-like object; got {type(file).__name__}."
    )
    raise TypeError(msg)


def unpack_7zarchive(archive, path, extra=None):
    """
    Function for registering with shutil.register_unpack_format().
    """
    if not is_7zfile(archive):
        raise ReadError(f"{archive} is not a 7zip file.")
    with SevenZipFile(archive) as arc:
        arc.extractall(path)


def pack_7zarchive(base_name, base_dir, owner=None, group=None, dry_run=None, logger=None):
    """
    Function for registering with shutil.register_archive_format().
    """
    target_name = f"{base_name}.7z"
    with SevenZipFile(target_name, mode="w") as archive:
        archive.writeall(path=base_dir)
    return target_name


class Worker:
    """
    Extract worker class to invoke handler.
    """

    def __init__(self, files, src_start: int, header, mp=False) -> None:
        self.target_filepath: dict[int, MemIO | pathlib.Path | None] = {}
        self.files = files
        self.src_start = src_start
        self.header = header
        self.current_file_index = len(self.files)
        self.last_file_index = len(self.files) - 1
        self.max_extract_size: int | None = None
        self._total_extracted: int = 0
        if mp:
            self.concurrent: type[Thread] | type[Process] = Process
        else:
            self.concurrent = Thread

    def extract(
        self,
        fp: IO[bytes],
        path: pathlib.Path | None,
        parallel: bool,
        skip_notarget=True,
        q=None,
    ) -> None:
        """Extract worker method to handle 7zip folder and decompress each files."""
        if hasattr(self.header, "main_streams") and self.header.main_streams is not None:
            src_end = self.src_start + self.header.main_streams.packinfo.packpositions[-1]
            numfolders = self.header.main_streams.unpackinfo.numfolders
            if numfolders == 1:
                self.extract_single(
                    fp,
                    self.files,
                    path,
                    self.src_start,
                    src_end,
                    q,
                    skip_notarget=skip_notarget,
                )
            else:
                folders = self.header.main_streams.unpackinfo.folders
                positions = self.header.main_streams.packinfo.packpositions
                empty_files = [f for f in self.files if f.emptystream]
                if not parallel:
                    self.extract_single(fp, empty_files, path, 0, 0, q)
                    for i in range(numfolders):
                        if skip_notarget:
                            if not any([self.target_filepath.get(f.id, None) for f in folders[i].files]):
                                continue
                        self.extract_single(
                            fp,
                            folders[i].files,
                            path,
                            self.src_start + positions[i],
                            self.src_start + positions[i + 1],
                            q,
                            skip_notarget=skip_notarget,
                        )
                else:
                    if getattr(fp, "name", None) is None:
                        raise InternalError("Caught unknown variable status error")
                    filename: str = getattr(fp, "name", "")  # do not become "" but it is for type check.
                    self.extract_single(open(filename, "rb"), empty_files, path, 0, 0, q)
                    concurrent_tasks = []
                    exc_q: queue.Queue = queue.Queue()
                    for i in range(numfolders):
                        if skip_notarget:
                            if not any([self.target_filepath.get(f.id, None) for f in folders[i].files]):
                                continue
                        p = self.concurrent(
                            target=self.extract_single,
                            args=(
                                filename,
                                folders[i].files,
                                path,
                                self.src_start + positions[i],
                                self.src_start + positions[i + 1],
                                q,
                                exc_q,
                                skip_notarget,
                            ),
                        )
                        p.start()
                        concurrent_tasks.append(p)
                    for p in concurrent_tasks:
                        p.join()
                    if exc_q.empty():
                        pass
                    else:
                        exc_info = exc_q.get()
                        raise exc_info[1].with_traceback(exc_info[2])
        else:
            empty_files = [f for f in self.files if f.emptystream]
            self.extract_single(fp, empty_files, path, 0, 0, q)

    def extract_single(
        self,
        fp: IO[bytes] | str,
        files,
        path,
        src_start: int,
        src_end: int,
        q: queue.Queue | None,
        exc_q: queue.Queue | None = None,
        skip_notarget=True,
    ) -> None:
        """
        Single thread extractor that takes file lists in single 7zip folder.
        """
        if files is None:
            return
        try:
            if isinstance(fp, str):
                fp = open(fp, "rb")
            fp.seek(src_start)
            self._extract_single(fp, files, path, src_end, q, skip_notarget)
        except Exception as e:
            if exc_q is None:
                raise e
            else:
                exc_tuple = sys.exc_info()
                exc_q.put(exc_tuple)

    def _extract_single(
        self,
        fp: IO[bytes],
        files,
        path,
        src_end: int,
        q: queue.Queue | None,
        skip_notarget=True,
    ) -> None:
        """
        Single thread extractor that takes file lists in single 7zip folder.
        this may raise exception.
        """
        just_check: list[ArchiveFile] = []
        for f in files:
            if q is not None:
                q.put(
                    (
                        "s",
                        str(f.filename),
                        str(f.compressed) if f.compressed is not None else "0",
                    )
                )
            fileish = self.target_filepath.get(f.id, None)
            if fileish is None:
                if not f.emptystream:
                    just_check.append(f)
            else:
                # delayed execution of crc check.
                self._check(fp, just_check, src_end)
                just_check = []
                if not isinstance(fileish, MemIO):
                    if not is_path_valid(fileish, path):
                        raise Bad7zFile(f"Specified path is bad: {fileish}")
                    fileish.parent.mkdir(parents=True, exist_ok=True)
                if not f.emptystream:
                    if f.is_junction and not isinstance(fileish, MemIO) and sys.platform == "win32":
                        with io.BytesIO() as ofp:
                            self.decompress(fp, f.folder, ofp, f.uncompressed, f.compressed, src_end, q)
                            dst: str = ofp.read().decode("utf-8")
                            if is_path_valid(fileish.parent.joinpath(dst), path):
                                # fileish.unlink(missing_ok=True) > py3.7
                                if fileish.exists():
                                    fileish.unlink()
                                if sys.platform == "win32":  # hint for mypy
                                    _winapi.CreateJunction(str(fileish), dst)  # noqa
                            else:
                                raise Bad7zFile("Junction point out of target directory.")
                    elif f.is_symlink and not isinstance(fileish, MemIO):
                        with io.BytesIO() as omfp:
                            self.decompress(fp, f.folder, omfp, f.uncompressed, f.compressed, src_end, q)
                            omfp.seek(0)
                            dst = omfp.read().decode("utf-8")
                            # check sym_target points inside an archive target?
                            if is_path_valid(fileish.parent.joinpath(dst), path):
                                sym_target = pathlib.Path(dst)
                                # fileish.unlink(missing_ok=True) > py3.7
                                if fileish.exists():
                                    fileish.unlink()
                                fileish.symlink_to(sym_target)
                            else:
                                raise Bad7zFile("Symlink point out of target directory.")
                    else:
                        with fileish.open(mode="wb") as obfp:
                            crc32 = self.decompress(fp, f.folder, obfp, f.uncompressed, f.compressed, src_end, q)
                            obfp.seek(0)
                            if f.crc32 is not None and crc32 != f.crc32:
                                raise CrcError(crc32, f.crc32, f.filename)
                else:
                    # just create empty file
                    if not isinstance(fileish, MemIO):
                        fileish.touch()
                    else:
                        with fileish.open() as ofp:
                            pass
            if q is not None:
                q.put(("e", str(f.filename), str(f.uncompressed)))
        if not skip_notarget:
            # delayed execution of crc check.
            self._check(fp, just_check, src_end)

    def _check(self, fp, check_target, src_end):
        """
        delayed execution of crc check.
        """
        for f in check_target:
            with NullIO() as ofp:
                crc32 = self.decompress(fp, f.folder, ofp, f.uncompressed, f.compressed, src_end)
            if f.crc32 is not None and crc32 != f.crc32:
                raise CrcError(crc32, f.crc32, f.filename)

    def decompress(
        self,
        fp: IO[bytes],
        folder,
        fq: IO[Any],
        size: int,
        compressed_size: int | None,
        src_end: int,
        q: queue.Queue | None = None,
    ) -> int:
        """
        decompressor wrapper called from extract method.

        :parameter fp: archive source file pointer
        :parameter folder: Folder object that have decompressor object.
        :parameter fq: output file pathlib.Path
        :parameter size: uncompressed size of target file.
        :parameter compressed_size: compressed size of target file.
        :parameter src_end: end position of the folder
        :parameter q: the queue for the reporter

        :returns None

        """
        assert folder is not None
        out_remaining = size
        max_block_size = get_memory_limit()
        crc32 = 0
        decompressor = folder.get_decompressor(compressed_size)
        previous_update_at = time.time()
        decompressed_bytes = 0
        while out_remaining > 0:
            tmp = decompressor.decompress(fp, min(out_remaining, max_block_size))
            if len(tmp) > 0:
                if self.max_extract_size is not None:
                    self._total_extracted += len(tmp)
                    if self._total_extracted > self.max_extract_size:
                        raise DecompressionBombError(
                            f"Extraction aborted: decompressed size {self._total_extracted} "
                            f"exceeds limit of {self.max_extract_size} bytes"
                        )
                out_remaining -= len(tmp)
                fq.write(tmp)
                crc32 = calculate_crc32(tmp, crc32)
            if q is not None:
                time_delta = time.time() - previous_update_at
                decompressed_bytes += len(tmp)
                if out_remaining <= 0 or time_delta >= 1:
                    q.put(("u", None, str(decompressed_bytes)))
                    previous_update_at += time_delta
                    decompressed_bytes = 0
            if out_remaining <= 0:
                break
        if fp.tell() >= src_end:
            if decompressor.crc is not None and not decompressor.check_crc():
                raise CrcError(decompressor.crc, decompressor.digest, None)
        return crc32

    def _find_link_target(self, target):
        """
        Find the target member of a symlink or hardlink member in the archive.
        """
        targetname: str = target.as_posix()
        linkname: str | pathlib.Path = readlink(targetname)
        # Check windows full path symlinks
        if str(linkname).startswith("\\\\?\\"):
            linkname = str(linkname)[4:]
        # normalize as posix style
        linkname = pathlib.Path(linkname).as_posix()
        member = None
        for j in range(len(self.files)):
            if linkname == self.files[j].origin.as_posix():
                # FIXME: when API user specify arcname, it will break
                member = os.path.relpath(linkname, os.path.dirname(targetname))
                break
        if member is None:
            member = linkname
        return member

    def _after_write(self, insize, foutsize, crc):
        self.header.main_streams.substreamsinfo.digestsdefined.append(True)
        self.header.main_streams.substreamsinfo.digests.append(crc)
        if self.header.main_streams.substreamsinfo.unpacksizes is None:
            self.header.main_streams.substreamsinfo.unpacksizes = [insize]
        else:
            self.header.main_streams.substreamsinfo.unpacksizes.append(insize)
        if self.header.main_streams.substreamsinfo.num_unpackstreams_folders is None:
            self.header.main_streams.substreamsinfo.num_unpackstreams_folders = [1]
        else:
            self.header.main_streams.substreamsinfo.num_unpackstreams_folders[-1] += 1
        return foutsize, crc

    def write(self, fp: IO[bytes], f, assym, folder):
        compressor = folder.get_compressor()
        if assym:
            link_target: str = self._find_link_target(f.origin)
            tgt: bytes = link_target.encode("utf-8")
            fd = io.BytesIO(tgt)
            insize, foutsize, crc = compressor.compress(fd, fp)
            fd.close()
        else:
            with f.origin.open(mode="rb") as fd:
                insize, foutsize, crc = compressor.compress(fd, fp)
        return self._after_write(insize, foutsize, crc)

    def writestr(self, fp: IO[bytes], f, folder):
        compressor = folder.get_compressor()
        insize, foutsize, crc = compressor.compress(f.data(), fp)
        return self._after_write(insize, foutsize, crc)

    def flush_archive(self, fp, folder):
        compressor = folder.get_compressor()
        foutsize = compressor.flush(fp)
        if len(self.files) > 0:
            if "maxsize" in self.header.files_info.files[self.last_file_index]:
                self.header.files_info.files[self.last_file_index]["maxsize"] += foutsize
            else:
                self.header.files_info.files[self.last_file_index]["maxsize"] = foutsize
        # Update size data in header
        self.header.main_streams.packinfo.numstreams += 1
        if self.header.main_streams.packinfo.enable_digests:
            self.header.main_streams.packinfo.crcs.append(compressor.digest)
            self.header.main_streams.packinfo.digestdefined.append(True)
        self.header.main_streams.packinfo.packsizes.append(compressor.packsize)
        folder.unpacksizes = compressor.unpacksizes

    def archive(self, fp: IO[bytes], files, folder, deref=False):
        """Run archive task for specified 7zip folder."""
        f = files[self.current_file_index]
        if f.has_strdata():
            foutsize, crc = self.writestr(fp, f, folder)
            self.header.files_info.files[self.current_file_index]["maxsize"] = foutsize
            self.header.files_info.files[self.current_file_index]["digest"] = crc
            self.last_file_index = self.current_file_index
        elif (f.is_symlink and not deref) or not f.emptystream:
            foutsize, crc = self.write(fp, f, (f.is_symlink and not deref), folder)
            self.header.files_info.files[self.current_file_index]["maxsize"] = foutsize
            self.header.files_info.files[self.current_file_index]["digest"] = crc
            self.last_file_index = self.current_file_index
        self.current_file_index += 1

    def register_filelike(self, id: int, fileish: MemIO | pathlib.Path | None) -> None:
        """register file-ish to worker."""
        self.target_filepath[id] = fileish

    def close(self):
        del self.header
        del self.files
        del self.concurrent



#
# p7zr library
#
# Copyright (c) 2019-2021 Hiroshi Miura <miurahr@linux.com>
# Copyright (c) 2004-2015 by Joachim Bauch, mail@joachim-bauch.de
# 7-Zip Copyright (C) 1999-2010 Igor Pavlov
# LZMA SDK Copyright (C) 1999-2010 Igor Pavlov
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
import binascii
import lzma
import platform
import sys

MAGIC_7Z = binascii.unhexlify("377abcaf271c")
FINISH_7Z = binascii.unhexlify("377abcaf271d")

COMMAND_HELP_STRING = """<Commands>
  c : Create archive with files
  i : Show information about supported formats
  l : List contents of archive
  t : Test integrity of archive
  x : eXtract files with full paths
"""


def is_64bit() -> bool:
    """check if running platform is 64bit python."""
    return sys.maxsize > (1 << 32)


def is_pypy369later() -> bool:
    """check if running platform is PyPY and python 3.6.9 and later"""
    return platform.python_implementation() == "PyPy" and sys.version_info >= (3, 6, 9)


def get_default_blocksize() -> int:
    """
    Return a safe buffer size for pass to decompress/compress modules.
    We check several conditions;

    1. 32bit python
      It sometimes fails with memory_error
      when decompress/compress large file (hundreds of MiB) on 32bit python.
      We reduce buffer size smaller to avoid memory_error.
      @see issue https://github.com/miurahr/py7zr/issues/370

    2. CPython
      CPython implementation of 3.7.5 fixed a lzma module bug
      that is not respect max_length parameter.
      When buffer size is larger than default of the module,
      max_length can be a value to lead the bug, so we set it
      lzma module default = 32768 bytes.
      @see BPO-21872: LZMA library sometimes fails to decompress a file
           https://bugs.python.org/issue21872
      @see https://github.com/miurahr/py7zr/issues/272

    3. PyPy
      PyPy 7.2 (python 3.6.9) fixed a lzma module's bug as same as
      CPython above.
      We set buffer size is as default size of the module to avoid the bugs.
      @see PyPy3-3090: lzma.LZMADecomporessor.decompress does not respect max_length
           https://foss.heptapod.net/pypy/pypy/-/issues/3090
      @see https://github.com/miurahr/py7zr/pull/114

    :return: recommended buffer size as int.
    """
    if is_64bit() and (is_pypy369later() or sys.version_info >= (3, 7, 5)):
        return 1048576
    else:
        return 32768


def get_memory_limit():
    """
    Get memory limit for allocating decompression chunk buffer.
    :return: allowed chunk size in bytes.
    """
    default_limit = int(128e6)
    if sys.platform.startswith("linux") or sys.platform.startswith("darwin") or sys.platform.startswith("openbsd"):
        import resource

        import psutil

        try:
            soft, _ = resource.getrlimit(resource.RLIMIT_DATA)
            if soft == -1:
                avmem = psutil.virtual_memory().available
                return min(default_limit, (avmem - int(256e6)) >> 2)
            else:
                return min(default_limit, (soft - int(256e6)) >> 2)
        except AttributeError:
            pass
    # fall back to default
    return default_limit


# Exposed constants
FILTER_LZMA = lzma.FILTER_LZMA1
FILTER_LZMA2 = lzma.FILTER_LZMA2
FILTER_DELTA = lzma.FILTER_DELTA
FILTER_ARM = lzma.FILTER_ARM
FILTER_ARMTHUMB = lzma.FILTER_ARMTHUMB
FILTER_IA64 = lzma.FILTER_IA64
FILTER_POWERPC = lzma.FILTER_POWERPC
FILTER_SPARC = lzma.FILTER_SPARC
FILTER_X86 = lzma.FILTER_X86
CHECK_CRC32 = lzma.CHECK_CRC32
CHECK_CRC64 = lzma.CHECK_CRC64
CHECK_SHA256 = lzma.CHECK_SHA256
CHECK_NONE = lzma.CHECK_NONE
CHECK_ID_MAX = lzma.CHECK_ID_MAX
CHECK_UNKNOWN = lzma.CHECK_UNKNOWN
PRESET_DEFAULT = lzma.PRESET_DEFAULT
PRESET_EXTREME = lzma.PRESET_EXTREME
FILTER_CRYPTO_AES256_SHA256 = 0x06F10701
FILTER_CRYPTO_ZIP = 0x06F10101
FILTER_CRYPTO_RAR29 = 0x06F10303

FILTER_BZIP2 = 0x31
FILTER_DEFLATE = 0x32
FILTER_COPY = 0x33
FILTER_ZSTD = 0x35
FILTER_PPMD = 0x36
FILTER_BROTLI = 0x37
FILTER_DEFLATE64 = 0x38


class Constant:
    """Constant base class."""

    def __setattr__(self, *_):
        pass


class DefaultFilters(Constant):
    """Default filter values."""

    ARCHIVE_FILTER = [{"id": FILTER_X86}, {"id": FILTER_LZMA2, "preset": 7 | PRESET_DEFAULT}]
    ENCODED_HEADER_FILTER = [{"id": FILTER_LZMA2, "preset": 7 | PRESET_DEFAULT}]
    ENCRYPTED_ARCHIVE_FILTER = [{"id": FILTER_LZMA2, "preset": 7 | PRESET_DEFAULT}, {"id": FILTER_CRYPTO_AES256_SHA256}]
    ENCRYPTED_HEADER_FILTER = [{"id": FILTER_CRYPTO_AES256_SHA256}]


DEFAULT_FILTERS = DefaultFilters()


class Property(Constant):
    """Hold 7zip property fixed values."""

    END = binascii.unhexlify("00")
    HEADER = binascii.unhexlify("01")
    ARCHIVE_PROPERTIES = binascii.unhexlify("02")
    ADDITIONAL_STREAMS_INFO = binascii.unhexlify("03")
    MAIN_STREAMS_INFO = binascii.unhexlify("04")
    FILES_INFO = binascii.unhexlify("05")
    PACK_INFO = binascii.unhexlify("06")
    UNPACK_INFO = binascii.unhexlify("07")
    SUBSTREAMS_INFO = binascii.unhexlify("08")
    SIZE = binascii.unhexlify("09")
    CRC = binascii.unhexlify("0a")
    FOLDER = binascii.unhexlify("0b")
    CODERS_UNPACK_SIZE = binascii.unhexlify("0c")
    NUM_UNPACK_STREAM = binascii.unhexlify("0d")
    EMPTY_STREAM = binascii.unhexlify("0e")
    EMPTY_FILE = binascii.unhexlify("0f")
    ANTI = binascii.unhexlify("10")
    NAME = binascii.unhexlify("11")
    CREATION_TIME = binascii.unhexlify("12")
    LAST_ACCESS_TIME = binascii.unhexlify("13")
    LAST_WRITE_TIME = binascii.unhexlify("14")
    ATTRIBUTES = binascii.unhexlify("15")
    COMMENT = binascii.unhexlify("16")
    ENCODED_HEADER = binascii.unhexlify("17")
    START_POS = binascii.unhexlify("18")
    DUMMY = binascii.unhexlify("19")


PROPERTY = Property()


class CompressionMethod(Constant):
    """Hold fixed values for method parameter."""

    COPY = binascii.unhexlify("00")
    DELTA = binascii.unhexlify("03")
    BCJ = binascii.unhexlify("04")
    PPC = binascii.unhexlify("05")
    IA64 = binascii.unhexlify("06")
    ARM = binascii.unhexlify("07")
    ARMT = binascii.unhexlify("08")
    SPARC = binascii.unhexlify("09")
    # SWAP = 02..
    SWAP2 = binascii.unhexlify("020302")
    SWAP4 = binascii.unhexlify("020304")
    # 7Z = 03..
    LZMA = binascii.unhexlify("030101")
    PPMD = binascii.unhexlify("030401")
    P7Z_BCJ = binascii.unhexlify("03030103")
    P7Z_BCJ2 = binascii.unhexlify("0303011B")
    BCJ_PPC = binascii.unhexlify("03030205")
    BCJ_IA64 = binascii.unhexlify("03030401")
    BCJ_ARM = binascii.unhexlify("03030501")
    BCJ_ARMT = binascii.unhexlify("03030701")
    BCJ_SPARC = binascii.unhexlify("03030805")
    LZMA2 = binascii.unhexlify("21")
    # MISC : 04..
    MISC_ZIP = binascii.unhexlify("0401")
    MISC_BZIP2 = binascii.unhexlify("040202")
    MISC_DEFLATE = binascii.unhexlify("040108")
    MISC_DEFLATE64 = binascii.unhexlify("040109")
    MISC_Z = binascii.unhexlify("0405")
    MISC_LZH = binascii.unhexlify("0406")
    NSIS_DEFLATE = binascii.unhexlify("040901")
    NSIS_BZIP2 = binascii.unhexlify("040902")
    #
    MISC_ZSTD = binascii.unhexlify("04f71101")
    MISC_BROTLI = binascii.unhexlify("04f71102")
    MISC_LZ4 = binascii.unhexlify("04f71104")
    MISC_LZS = binascii.unhexlify("04f71105")
    MISC_LIZARD = binascii.unhexlify("04f71106")
    # CRYPTO 06..
    CRYPT_ZIPCRYPT = binascii.unhexlify("06f10101")
    CRYPT_RAR29AES = binascii.unhexlify("06f10303")
    CRYPT_AES256_SHA256 = binascii.unhexlify("06f10701")


COMPRESSION_METHOD = CompressionMethod()



#!/usr/bin/python -u
#
# p7zr library
#
# Copyright (c) 2019-2024 Hiroshi Miura <miurahr@linux.com>
# Copyright (c) 2004-2015 by Joachim Bauch, mail@joachim-bauch.de
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
from __future__ import annotations

import ctypes
import hashlib
import os
import pathlib
import platform
import posixpath
import re
import sys
import time as _time
import zlib
from datetime import datetime, timedelta, timezone, tzinfo
from typing import TYPE_CHECKING

from py7zr import Bad7zFile
from py7zr.win32compat import is_windows_native_python, is_windows_unc_path

if TYPE_CHECKING:
    from typing_extensions import Self

# String used at the beginning of relative paths
RELATIVE_PATH_MARKER = "./"


def calculate_crc32(data: bytes, value: int = 0, blocksize: int = 1024 * 1024) -> int:
    """Calculate CRC32 of strings with arbitrary lengths."""
    if len(data) <= blocksize:
        value = zlib.crc32(data, value)
    else:
        length = len(data)
        pos = blocksize
        value = zlib.crc32(data[:pos], value)
        while pos < length:
            value = zlib.crc32(data[pos : pos + blocksize], value)
            pos += blocksize
    return value & 0xFFFFFFFF


def _get_hash(digest: str):
    if digest not in hashlib.algorithms_available:
        raise ValueError("Unknown digest method for password protection.")
    if digest == "sha256":
        return hashlib.sha256()
    return hashlib.new(digest)


def _calculate_key1(password: bytes, cycles: int, salt: bytes, digest: str) -> bytes:
    """Calculate 7zip AES encryption key. Base implementation."""
    assert cycles <= 0x3F
    if cycles == 0x3F:
        ba = bytearray(salt + password + bytes(32))
        key: bytes = bytes(ba[:32])
    else:
        rounds = 1 << cycles
        m = _get_hash(digest)
        for round in range(rounds):
            m.update(salt + password + round.to_bytes(8, byteorder="little", signed=False))
        key = m.digest()[:32]
    return key


def _calculate_key2(password: bytes, cycles: int, salt: bytes, digest: str):
    """Calculate 7zip AES encryption key.
    It uses ctypes and memoryview buffer and zero-copy technology on Python."""
    assert cycles <= 0x3F
    if cycles == 0x3F:
        key: bytes = bytes(bytearray(salt + password + bytes(32))[:32])
    else:
        rounds = 1 << cycles
        m = _get_hash(digest)
        length = len(salt) + len(password)

        class RoundBuf(ctypes.LittleEndianStructure):
            _pack_ = 1
            _fields_ = [
                ("saltpassword", ctypes.c_ubyte * length),
                ("round", ctypes.c_uint64),
            ]

        buf = RoundBuf()
        for i, c in enumerate(salt + password):
            buf.saltpassword[i] = c
        buf.round = 0
        mv = memoryview(buf)
        while buf.round < rounds:
            m.update(mv)
            buf.round += 1
        key = m.digest()[:32]
    return key


def _calculate_key3(password: bytes, cycles: int, salt: bytes, digest: str) -> bytes:
    """Calculate 7zip AES encryption key.
    Concat values in order to reduce number of calls of Hash.update()."""
    assert cycles <= 0x3F
    if cycles == 0x3F:
        ba = bytearray(salt + password + bytes(32))
        key: bytes = bytes(ba[:32])
    else:
        cat_cycle = 6
        if cycles > cat_cycle:
            rounds = 1 << cat_cycle
            stages = 1 << (cycles - cat_cycle)
        else:
            rounds = 1 << cycles
            stages = 1 << 0
        m = _get_hash(digest)
        saltpassword = salt + password
        s = 0  # type: int  # (0..stages) * rounds
        if platform.python_implementation() == "PyPy":
            for _ in range(stages):
                m.update(
                    memoryview(
                        b"".join(
                            [saltpassword + (s + i).to_bytes(8, byteorder="little", signed=False) for i in range(rounds)]
                        )
                    )
                )
                s += rounds
        else:
            for _ in range(stages):
                m.update(
                    b"".join([saltpassword + (s + i).to_bytes(8, byteorder="little", signed=False) for i in range(rounds)])
                )
                s += rounds
        key = m.digest()[:32]

    return key


if platform.python_implementation() == "PyPy" or sys.version_info > (3, 6):
    calculate_key = _calculate_key3
else:
    calculate_key = _calculate_key2  # it is faster when CPython 3.6.x


def filetime_to_dt(ft):
    """Convert Windows NTFS file time into python datetime object."""
    EPOCH_AS_FILETIME = 116444736000000000
    us = (ft - EPOCH_AS_FILETIME) // 10
    return datetime(1970, 1, 1, tzinfo=timezone.utc) + timedelta(microseconds=us)


ZERO = timedelta(0)
HOUR = timedelta(hours=1)
SECOND = timedelta(seconds=1)

# A class capturing the platform's idea of local time.
# (May result in wrong values on historical times in
#  timezones where UTC offset and/or the DST rules had
#  changed in the past.)

STDOFFSET = timedelta(seconds=-_time.timezone)
if _time.daylight:
    DSTOFFSET = timedelta(seconds=-_time.altzone)
else:
    DSTOFFSET = STDOFFSET

DSTDIFF = DSTOFFSET - STDOFFSET


class LocalTimezone(tzinfo):
    def fromutc(self, dt):
        assert dt.tzinfo is self
        stamp = (dt - datetime(1970, 1, 1, tzinfo=self)) // SECOND
        args = _time.localtime(stamp)[:6]
        # dst_diff = DSTDIFF // SECOND
        # Detect fold
        # fold = args == _time.localtime(stamp - dst_diff)
        return datetime(*args, microsecond=dt.microsecond, tzinfo=self)

    def utcoffset(self, dt):
        if self._isdst(dt):
            return DSTOFFSET
        else:
            return STDOFFSET

    def dst(self, dt):
        if self._isdst(dt):
            return DSTDIFF
        else:
            return ZERO

    def tzname(self, dt):
        return _time.tzname[self._isdst(dt)]

    def _isdst(self, dt):
        tt = (
            dt.year,
            dt.month,
            dt.day,
            dt.hour,
            dt.minute,
            dt.second,
            dt.weekday(),
            0,
            0,
        )
        stamp = _time.mktime(tt)
        tt = _time.localtime(stamp)
        return tt.tm_isdst > 0


Local = LocalTimezone()
TIMESTAMP_ADJUST = -11644473600


class UTC(tzinfo):
    """UTC"""

    def utcoffset(self, dt):
        return ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return ZERO

    def _call__(self):
        return self


class ArchiveTimestamp(int):
    """Windows FILETIME timestamp."""

    def __repr__(self):
        return f"{type(self).__name__}({self})"

    def __index__(self):
        return self.__int__()

    def totimestamp(self) -> float:
        """Convert 7z FILETIME to Python timestamp."""
        # FILETIME is 100-nanosecond intervals since 1601/01/01 (UTC)
        return (self / 10000000.0) + TIMESTAMP_ADJUST

    def as_datetime(self) -> datetime:
        """Convert FILETIME to Python datetime object."""
        return datetime.fromtimestamp(self.totimestamp(), UTC())

    @classmethod
    def from_datetime(cls, val: float | int) -> Self:
        return cls((val - TIMESTAMP_ADJUST) * 10000000.0)

    @classmethod
    def from_now(cls) -> Self:
        return cls((_time.time() - TIMESTAMP_ADJUST) * 10000000.0)


def islink(path: str | pathlib.Path) -> bool:
    """
    Cross-platform islink implementation.
    Support Windows NT symbolic links and reparse points.
    """
    return os.path.islink(path)


def readlink(path: str | pathlib.Path, *, dir_fd=None) -> str | pathlib.Path:
    """
    Cross-platform compat implementation of os.readlink and Path.readlink().
    Support Windows NT symbolic links and reparse points.
    When called with path argument as pathlike(str), return result as a pathlike(str).
    When called with Path object, return also Path object.
    When called with path argument as bytes, return result as a bytes.
    """
    if not os.path.exists(str(path)):
        raise OSError(22, "Invalid argument", path)

    if isinstance(path, pathlib.Path) and dir_fd is None:
        return path.readlink()
    else:
        return os.readlink(path, dir_fd=dir_fd)


def remove_relative_path_marker(path: str) -> str:
    """
    Removes './' from the beginning of a path-like string
    """
    processed_path = path

    if path.startswith(RELATIVE_PATH_MARKER):
        processed_path = path[len(RELATIVE_PATH_MARKER) :]

    return processed_path


def remove_trailing_slash(path: str) -> str:
    """
    Removes '/' from the end of a path-like string
    """
    if path.endswith(posixpath.sep):
        return path[:-1]
    return path


def canonical_path(target: pathlib.Path) -> pathlib.Path:
    """Return a canonical path of target argument."""
    stack: list[str] = []
    anchor = target.anchor  # preserves POSIX '/' and Windows 'C:\\' or UNC share
    for p in target.parts:
        if p != ".." or len(stack) == 0:
            stack.append(p)
            continue
        # treat '..'
        if stack[-1] == "..":
            stack.append(p)  # '../' + '../' -> '../../'
        elif anchor and stack[-1] == anchor:
            pass  # don't walk above the anchor/drive/share
        else:
            stack.pop()  # 'foo/boo/' + '..' -> 'foo/'
    return pathlib.Path(*stack)


def is_relative_to(my: pathlib.Path, *other) -> bool:
    """Return True when path is relative to other path, otherwise False."""
    base = canonical_path(*other)
    try:
        # resolve symlinks to catch Zip-Slip style vulnerability
        my_resolved = my.resolve(strict=False)
        base_resolved = base.resolve(strict=False)
        if my_resolved == base_resolved:
            # don't allow paths to be equal to base
            return False
        my_resolved.relative_to(base_resolved)
    except (ValueError, RuntimeError):
        return False
    return True


def get_sanitized_output_path(fname: str, path: pathlib.Path | None) -> pathlib.Path:
    """
    check f.filename has invalid directory traversals
    When condition is not satisfied, raise Bad7zFile
    """
    # Check for Windows drive letter (e.g., C:\, D:\)
    # This is a security risk even on non-Windows systems
    if re.match(r"^[A-Za-z]:", fname):
        raise Bad7zFile(f"Specified path is bad: {fname}")

    if fname.startswith("/"):
        fname = fname.lstrip("/")

    if path is None:
        target_path = canonical_path(pathlib.Path.cwd().joinpath(fname))
        if is_relative_to(target_path, pathlib.Path.cwd()):
            return pathlib.Path(remove_relative_path_marker(fname))
    else:
        outfile = canonical_path(path.joinpath(remove_relative_path_marker(fname)))
        if is_relative_to(outfile, path):
            return pathlib.Path(outfile)
    raise Bad7zFile(f"Specified path is bad: {fname}")


def check_archive_path(arcname: str) -> bool:
    """
    Check arcname argument is valid for archive.
    It should not be absolute, if so it returns False.
    It should not be evil traversal attack path.
    Otherwise, returns True.
    """
    if pathlib.Path(arcname).is_absolute():
        return False
    # test against dummy parent path
    if sys.platform == "win32":
        path = pathlib.Path("C:/foo/boo/fuga/hoge/a90sufoiasj09/dafj08sajfa/")
    else:
        path = pathlib.Path("/foo/boo/fuga/hoge/a90sufoiasj09/dafj08sajfa/")
    return is_path_valid(path.joinpath(arcname), path)


def is_path_valid(target: pathlib.Path, parent: pathlib.Path | None) -> bool:
    """
    Check if target path is valid against parent path.
    It returns False when resolved target path has '..', or point out of a parent path.
    Otherwise, returns True.
    """
    if parent is None:
        parent = pathlib.Path.cwd()
    cpath = canonical_path(target)
    if ".." in cpath.parts:
        return False
    if parent.is_absolute():
        return is_relative_to(cpath, parent)
    else:
        return is_relative_to(cpath, pathlib.Path.cwd().joinpath(parent))


def check_win32_file_namespace(pathname: pathlib.Path) -> pathlib.Path:
    # When python on Windows and not python on Cygwin,
    # Add win32 file namespace to exceed Microsoft Windows
    # path length limitation to 260 bytes
    # ref.
    # https://docs.microsoft.com/en-us/windows/win32/fileio/naming-a-file
    # In editions of Windows before Windows 10 version 1607,
    # the maximum length for a path is MAX_PATH, which is defined as
    # 260 characters. In later versions of Windows, changing a registry key
    # or select option when python installation is required to remove the limit.
    if is_windows_native_python() and pathname.is_absolute() and not is_windows_unc_path(pathname):
        pathname = pathlib.WindowsPath("\\\\?\\" + str(pathname))
    return pathname



#!/usr/bin/env python
#
#    Pure python p7zr implementation
#    Copyright (C) 2019-2021 Hiroshi Miura
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import argparse
import getpass
import inspect
import io
import lzma
import os
import pathlib
import platform
import re
import shutil
import sys
from lzma import CHECK_CRC64, CHECK_SHA256, is_check_supported
from typing import Any

import multivolumefile
import texttable  # type: ignore

import py7zr
from py7zr.callbacks import ExtractCallback
from py7zr.compressor import SupportedMethods
from py7zr.helpers import Local
from py7zr.properties import COMMAND_HELP_STRING


class CliExtractCallback(ExtractCallback):
    def __init__(self, total_bytes, ofd=sys.stdout):
        self.ofd = ofd
        self.archive_total = total_bytes
        self.total_bytes = 0
        self.columns, _ = shutil.get_terminal_size(fallback=(80, 24))
        self.pwidth = 0

    def report_start_preparation(self):
        pass

    def report_start(self, processing_file_path, processing_bytes):
        self.ofd.write(f"- {processing_file_path}")
        self.pwidth += len(processing_file_path) + 2

    def report_update(self, decompressed_bytes):
        pass

    def report_end(self, processing_file_path, wrote_bytes):
        self.total_bytes += int(wrote_bytes)
        plest = self.columns - self.pwidth
        progress = self.total_bytes / self.archive_total
        msg = f"({progress:.0%})\n"
        if plest - len(msg) > 0:
            self.ofd.write(msg.rjust(plest))
        else:
            self.ofd.write(msg)
        self.pwidth = 0

    def report_postprocess(self):
        pass

    def report_warning(self, message):
        pass


class Cli:
    dunits = {
        "b": 1,
        "B": 1,
        "k": 1024,
        "K": 1024,
        "m": 1024 * 1024,
        "M": 1024 * 1024,
        "g": 1024 * 1024 * 1024,
        "G": 1024 * 1024 * 1024,
    }

    def __init__(self):
        self.parser = self._create_parser()
        self.unit_pattern = re.compile(r"^([0-9]+)([bkmg]?)$", re.IGNORECASE)

    def run(self, arg: Any | None = None) -> int:
        args = self.parser.parse_args(arg)
        if args.version:
            return self.show_version()
        return args.func(args)

    def _create_parser(self):
        parser = argparse.ArgumentParser(
            prog="py7zr",
            description="py7zr",
            formatter_class=argparse.RawTextHelpFormatter,
            add_help=True,
        )
        subparsers = parser.add_subparsers(title="subcommands", help=COMMAND_HELP_STRING)
        list_parser = subparsers.add_parser("l")
        list_parser.set_defaults(func=self.run_list)
        list_parser.add_argument("arcfile", type=pathlib.Path, help="7z archive file")
        list_parser.add_argument("--verbose", action="store_true", help="verbose output")
        extract_parser = subparsers.add_parser("x")
        extract_parser.set_defaults(func=self.run_extract)
        extract_parser.add_argument("arcfile", type=pathlib.Path, help="7z archive file")
        extract_parser.add_argument("odir", nargs="?", help="output directory")
        extract_parser.add_argument("--files", nargs="+", help="List of archive members to extract")
        extract_parser.add_argument(
            "-P",
            "--password",
            action="store_true",
            help="Password protected archive(you will be asked a password).",
        )
        extract_parser.add_argument("--verbose", action="store_true", help="verbose output")
        create_parser = subparsers.add_parser("c")
        create_parser.set_defaults(func=self.run_create)
        create_parser.add_argument("arcfile", help="7z archive file")
        create_parser.add_argument("filenames", nargs="+", help="filenames to archive")
        create_parser.add_argument("-v", "--volume", nargs=1, help="Create volumes.")
        create_parser.add_argument(
            "-P",
            "--password",
            action="store_true",
            help="Password protected archive(you will be asked a password).",
        )
        append_parser = subparsers.add_parser("a")
        append_parser.set_defaults(func=self.run_append)
        append_parser.add_argument("arcfile", help="7z archive file")
        append_parser.add_argument("filenames", nargs="+", help="filenames to archive")
        test_parser = subparsers.add_parser("t")
        test_parser.set_defaults(func=self.run_test)
        test_parser.add_argument("arcfile", help="7z archive file")
        info_parser = subparsers.add_parser("i")
        info_parser.set_defaults(func=self.run_info)
        parser.add_argument("--version", action="store_true", help="Show version")
        parser.set_defaults(func=self.show_help)
        return parser

    def show_version(self):
        print(self._get_version())

    @staticmethod
    def _get_version():
        s = inspect.stack()
        _module = inspect.getmodule(s[0][0])
        module_name = _module.__name__ if _module is not None else "unknown"
        py_version = platform.python_version()
        py_impl = platform.python_implementation()
        py_build = platform.python_compiler()
        return (
            f"{module_name} Version {py7zr.__version__} : {py7zr.__copyright__} (Python {py_version} [{py_impl} {py_build}])"
        )

    def show_help(self, args):
        self.show_version()
        self.parser.print_help()
        return 0

    def run_info(self, args):
        self.show_version()
        print("\nFormats:")
        table = texttable.Texttable()
        table.set_deco(texttable.Texttable.HEADER)
        table.set_cols_dtype(["t", "t"])
        table.set_cols_align(["l", "r"])
        for f in SupportedMethods.formats:
            m = "".join(f" {x:02x}" for x in f["magic"])
            table.add_row([str(f["name"]), m])
        print(table.draw())
        print("\nCodecs and hashes:")
        table = texttable.Texttable()
        table.set_deco(texttable.Texttable.HEADER)
        table.set_cols_dtype(["t", "t"])
        table.set_cols_align(["l", "r"])
        for c in SupportedMethods.methods:
            method_id: bytes = c["id"]
            m = "".join(f"{x:02x}" for x in method_id)
            method_name: str = c["name"]
            table.add_row([m, method_name])
        table.add_row(["0", "CRC32"])
        if is_check_supported(CHECK_SHA256):
            table.add_row(["0", "SHA256"])
        if is_check_supported(CHECK_CRC64):
            table.add_row(["0", "CRC64"])
        print(table.draw())

    def run_list(self, args):
        """Print a table of contents to file."""
        target = args.arcfile
        verbose = args.verbose
        if re.fullmatch(r"[.]0+1?", target.suffix):
            mv_target = pathlib.Path(target.parent, target.stem)
            ext_start = int(target.suffix[-1])
            with multivolumefile.MultiVolume(
                mv_target, mode="rb", ext_digits=len(target.suffix) - 1, ext_start=ext_start
            ) as mvf:
                setattr(mvf, "name", str(mv_target))
                return self._run_list(mvf, verbose)
        else:
            return self._run_list(target, verbose)

    def _run_list(self, target, verbose):
        if not py7zr.is_7zfile(target):
            print("not a 7z file")
            return 1
        with py7zr.SevenZipFile(target, "r") as a:
            file = sys.stdout
            archive_info = a.archiveinfo()
            archive_list = a.list()
            if verbose:
                if isinstance(target, io.FileIO) or isinstance(target, multivolumefile.MultiVolume):
                    file_name: str = target.name  # type: ignore
                else:
                    file_name = str(target)
                file.write(f"Listing archive: {file_name}\n")
                file.write("--\n")
                file.write(f"Path = {archive_info.filename}\n")
                file.write("Type = 7z\n")
                fstat = archive_info.stat
                file.write(f"Physical Size = {fstat.st_size}\n")
                file.write(f"Headers Size = {archive_info.header_size}\n")
                file.write("Method = {}\n".format(", ".join(archive_info.method_names)))
                if archive_info.solid:
                    file.write("Solid = {}\n".format("+"))
                else:
                    file.write("Solid = {}\n".format("-"))
                file.write(f"Blocks = {archive_info.blocks}\n")
                file.write("\n")
            file.write(
                "total %d files and directories in %sarchive\n"
                % (len(archive_list), (archive_info.solid and "solid ") or "")
            )
            file.write("   Date      Time    Attr         Size   Compressed  Name\n")
            file.write("------------------- ----- ------------ ------------  ------------------------\n")
            for f in archive_list:
                if f.creationtime is not None:
                    lastwritedate = f.creationtime.astimezone(Local).strftime("%Y-%m-%d")
                    lastwritetime = f.creationtime.astimezone(Local).strftime("%H:%M:%S")
                else:
                    lastwritedate = "         "
                    lastwritetime = "         "
                if f.is_directory:
                    attrib = "D..."
                else:
                    attrib = "...."
                if f.archivable:
                    attrib += "A"
                else:
                    attrib += "."
                if f.is_directory:
                    extra = "           0 "
                elif f.compressed is None:
                    extra = "             "
                else:
                    extra = "%12d " % (f.compressed)
                file.write(
                    "%s %s %s %12d %s %s\n"
                    % (
                        lastwritedate,
                        lastwritetime,
                        attrib,
                        f.uncompressed,
                        extra,
                        f.filename,
                    )
                )
            file.write("------------------- ----- ------------ ------------  ------------------------\n")

        return 0

    @staticmethod
    def print_archiveinfo(archive, file):
        file.write("--\n")
        file.write(f"Path = {archive.filename}\n")
        file.write("Type = 7z\n")
        fstat = os.stat(archive.filename)
        file.write(f"Physical Size = {fstat.st_size}\n")
        file.write(f"Headers Size = {archive.header.size}\n")
        file.write("Method = {}\n".format(", ".join(archive._get_method_names())))
        if archive._is_solid():
            file.write("Solid = {}\n".format("+"))
        else:
            file.write("Solid = {}\n".format("-"))
        file.write(f"Blocks = {len(archive.header.main_streams.unpackinfo.folders)}\n")

    def run_test(self, args):
        target = args.arcfile
        if not py7zr.is_7zfile(target):
            print("not a 7z file")
            return 1
        with open(target, "rb") as f:
            try:
                a = py7zr.SevenZipFile(f)
                file = sys.stdout
                file.write(f"Testing archive: {a.filename}\n")
                self.print_archiveinfo(archive=a, file=file)
                file.write("\n")
                if a.testzip() is None:
                    file.write("Everything is Ok\n")
                    return 0
                else:
                    file.write("Bad 7zip file\n")
                    return 1
            except py7zr.exceptions.Bad7zFile:
                print("Header is corrupted. Cannot read as 7z file.")
                return 1
            except py7zr.exceptions.PasswordRequired:
                print("The archive is encrypted but password is not given. FAILED.")
                return 1

    def run_extract(self, args: argparse.Namespace) -> int:
        target = args.arcfile
        verbose = args.verbose
        if not py7zr.is_7zfile(target):
            print("not a 7z file")
            return 1
        if not args.password:
            password: str | None = None
        else:
            try:
                password = getpass.getpass()
            except getpass.GetPassWarning:
                sys.stderr.write("Warning: your password may be shown.\n")
                return 1
        try:
            a = py7zr.SevenZipFile(target, "r", password=password)
        except py7zr.exceptions.Bad7zFile:
            print("Header is corrupted. Cannot read as 7z file.")
            return 1
        except py7zr.exceptions.PasswordRequired:
            print("The archive is encrypted, but password is not given. ABORT.")
            return 1
        except lzma.LZMAError:
            if password is None:
                print("The archive is corrupted. ABORT.")
            else:
                print("The archive is corrupted, or password is wrong. ABORT.")
            return 1

        cb: ExtractCallback | None = None
        if verbose:
            archive_info = a.archiveinfo()
            cb = CliExtractCallback(total_bytes=archive_info.uncompressed, ofd=sys.stderr)
        try:
            if args.files:
                a.extract(
                    path=args.odir,
                    callback=cb,
                    targets=args.files,
                    # Allow to provide directories and extract all files in them.
                    recursive=True,
                )
            else:
                a.extractall(path=args.odir, callback=cb)
        except py7zr.exceptions.UnsupportedCompressionMethodError:
            print("Unsupported compression method is used in archive. ABORT.")
            return 1
        except py7zr.exceptions.DecompressionError:
            print("Error has been occurred during decompression. ABORT.")
            return 1
        except py7zr.exceptions.PasswordRequired:
            print("The archive is encrypted, but password is not given. ABORT.")
            return 1
        except lzma.LZMAError:
            if password is None:
                print("The archive is corrupted. ABORT.")
            else:
                print("The archive is corrupted, or password is wrong. ABORT.")
            return 1
        else:
            return 0

    def _check_volumesize_valid(self, size: str) -> bool:
        if self.unit_pattern.match(size):
            return True
        else:
            return False

    def _volumesize_unitconv(self, size: str) -> int:
        m = self.unit_pattern.match(size)
        if m is not None:
            num = m.group(1)
            unit = m.group(2)
            return int(num) if unit is None else int(num) * self.dunits[unit]
        else:
            return -1

    def run_create(self, args):
        sztarget: str = args.arcfile
        filenames: list[str] = args.filenames
        volume_size = args.volume[0] if getattr(args, "volume", None) is not None else None
        if volume_size is not None and not self._check_volumesize_valid(volume_size):
            sys.stderr.write("Error: Specified volume size is invalid.\n")
            self.show_help(args)
            exit(1)
        if not sztarget.endswith(".7z"):
            sztarget += ".7z"
        target = pathlib.Path(sztarget)
        if target.exists():
            sys.stderr.write("Archive file exists!\n")
            self.show_help(args)
            exit(1)
        if not args.password:
            password: str | None = None
        else:
            try:
                password = getpass.getpass()
            except getpass.GetPassWarning:
                sys.stderr.write("Warning: your password may be shown.\n")
                return 1
        if volume_size is None:
            with py7zr.SevenZipFile(target, "w", password=password) as szf:
                for path in filenames:
                    src = pathlib.Path(path)
                    if src.is_dir():
                        szf.writeall(src)
                    else:
                        szf.write(src)
            return 0
        else:
            size = self._volumesize_unitconv(volume_size)
            with multivolumefile.MultiVolume(target, mode="wb", volume=size, ext_digits=4) as mvf:
                with py7zr.SevenZipFile(mvf, "w", password=password) as szf:
                    for path in filenames:
                        src = pathlib.Path(path)
                        if src.is_dir():
                            szf.writeall(src)
                        else:
                            szf.write(src)
            return 0

    def run_append(self, args):
        sztarget: str = args.arcfile
        filenames: list[str] = args.filenames
        if not sztarget.endswith(".7z"):
            sys.stderr.write("Error: specified archive file is invalid.")
            self.show_help(args)
            exit(1)
        target = pathlib.Path(sztarget)
        if not target.exists():
            sys.stderr.write("Archive file does not exists!\n")
            self.show_help(args)
            exit(1)
        with py7zr.SevenZipFile(target, "a") as szf:
            for path in filenames:
                src = pathlib.Path(path)
                if src.is_dir():
                    szf.writeall(src)
                else:
                    szf.write(src)
        return 0



#
# p7zr library
#
# Copyright (c) 2019 Hiroshi Miura <miurahr@linux.com>
# Copyright (c) 2004-2015 by Joachim Bauch, mail@joachim-bauch.de
# 7-Zip Copyright (C) 1999-2010 Igor Pavlov
# LZMA SDK Copyright (C) 1999-2010 Igor Pavlov
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


from __future__ import annotations

from typing import Any


class ArchiveError(Exception):
    """Base class for exceptions."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class Bad7zFile(ArchiveError):
    pass


class CrcError(ArchiveError):
    """Exception raised for CRC error when decompression.

    Attributes:
      expected -- expected CRC bytes
      actual -- actual CRC data
      filename -- filename that has CRC error
    """

    def __init__(self, expected: Any, actual: Any, filename: Any) -> None:
        super().__init__(expected, actual, filename)
        self.expected = expected
        self.actual = actual
        self.filename = filename


class UnsupportedCompressionMethodError(ArchiveError):
    """Exception raised for unsupported compression parameter given.

    Attributes:
      data -- unknown property data
      message -- explanation of error
    """

    def __init__(self, data: Any, message: Any) -> None:
        super().__init__(data, message)
        self.data = data
        self.message = message


class DecompressionError(ArchiveError):
    pass


class InternalError(ArchiveError):
    pass


class PasswordRequired(Exception):
    pass


class DecompressionBombError(ArchiveError):
    """Exception raised when decompressed size exceeds the configured limit."""

    pass


class AbsolutePathError(Exception):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
